#include "datasetviewer.h"
#include "ui_datasetviewer.h"
#include "qwt_plot_grid.h"
#include "qwt_scale_engine.h"
#include "qwt_plot_marker.h"
#include "reduction/reductionservice.h"

DatasetViewer::DatasetViewer(InputDataset *dataset, QWidget *parent) :
    QDialog(parent),
    ui(new Ui::DatasetViewer)
{
	this->dataset = dataset;
	hasReverse = dataset->isReduced;
    ui->setupUi(this);

	ui->reversePlot->setVisible(false);

	int nSamples = dataset->nSamples;
	ui->sampleNumberSpinBox->setMinimum(0);
	ui->sampleNumberSpinBox->setMaximum(nSamples-1);

	ui->sampleNumberSlider->setMinimum(0);
	ui->sampleNumberSlider->setMaximum(nSamples-1);

	ui->divNsamples->setText("/ " + QString::number(nSamples));
	ui->nSamples->setText(QString::number(nSamples));
	ui->nDimensions->setText(QString::number(dataset->nDimensions));
	ui->nChannels->setText(QString::number(dataset->nChannels));

	connect(ui->sampleNumberSpinBox, SIGNAL(valueChanged(int)), this, SLOT(setSample(int)));
	connect(ui->sampleNumberSlider, SIGNAL(valueChanged(int)), this, SLOT(setSample(int)));
	connect(this, SIGNAL(indexChanged(int)), ui->sampleNumberSlider, SLOT(setValue(int)));
	connect(this, SIGNAL(indexChanged(int)), ui->sampleNumberSpinBox, SLOT(setValue(int)));

	curve = new QwtPlotCurve();
	xrange = new double[dataset->nDimensions];
	yrange = new double[dataset->nDimensions];
	plot = ui->plot;
	preparePlot(plot, dataset, curve, xrange);

	if(hasReverse) {
		Reduction *red = dataset->reduceClass;
		reverseCurve = new QwtPlotCurve();
		reverseDataset = red->reverse(dataset);
		reverseXrange = new double[reverseDataset->nDimensions];
		reverseYrange = new double[reverseDataset->nDimensions];
		reversePlot = ui->reversePlot;
		preparePlot(reversePlot, reverseDataset, reverseCurve, reverseXrange);
	} else {
		ui->pushButton->setVisible(false);
	}
	prevIndex = -1;

	showReverse = false;
	setSample(0);

}

void DatasetViewer::preparePlot(QwtPlot *plot, InputDataset *dataset, QwtPlotCurve *curve, double *xrange) {
	plot->setCanvasBackground(QColor(Qt::white));
	curve->setRenderHint(QwtPlotItem::RenderAntialiased);
	curve->setPen(QPen(QColor(Qt::blue)));
	curve->attach(plot);

	int nDimensions = dataset->nDimensions;
	for(int i=0;i<nDimensions;i++) {
		xrange[i] = i;
	}

	// Find max and min
	double maxY = -1e200;
	double minY = 1e200;
	std::vector<std::vector<double> >::iterator it;
	std::vector<double>::iterator itt;

	for(it = dataset->data.begin();it!=dataset->data.end();it++) {
		for(itt= it->begin();itt != it->end();itt++) {
			double val = *itt;
			if(val > maxY)
				maxY = val;
			if(val < minY)
				minY = val;
		}
	}

	plot->setAxisScale(QwtPlot::yLeft, minY, maxY);


	double interval;
	if((nDimensions == 256 || nDimensions == 200)) { //TODO could be made general
		interval = 50;
	} else if(dataset->nChannels > 1) {
		interval = nDimensions / dataset->nChannels;
	}

	for(unsigned int i=1;i<dataset->nChannels; i++) {
		QwtPlotMarker *mrk = new QwtPlotMarker();
		mrk->setLineStyle(QwtPlotMarker::VLine);
		mrk->setLinePen(QPen(Qt::red, 0, Qt::DashLine));
		mrk->setXValue(i*interval - 0.5);
		mrk->attach(plot);
		// TODO grey behind interval or something
	}


	QwtPlotMarker *zeroAxis = new QwtPlotMarker();
	zeroAxis->setLineStyle(QwtPlotMarker::HLine);
	zeroAxis->setLinePen(QPen(Qt::red));
	zeroAxis->setYValue(0);
	zeroAxis->attach(plot);
}

DatasetViewer::~DatasetViewer()
{
	delete ui;
	delete[] xrange;
	delete[] yrange;
	delete curve;
	//delete plot;

	if(hasReverse) {
		delete reverseDataset;
		delete[] reverseXrange;
		delete[] reverseYrange;
		delete reverseCurve;
		//delete reversePlot;
	}


}

void DatasetViewer::setSample(int sampleIndex) {
	if(prevIndex == sampleIndex)
		return;
	prevIndex = sampleIndex;
	emit(indexChanged(sampleIndex));

	if(!showReverse) {
		std::copy(dataset->data[sampleIndex].begin(), dataset->data[sampleIndex].end(), yrange);
		curve->setRawSamples(xrange, yrange, dataset->nDimensions);
		plot->replot();
	} else {
		std::copy(reverseDataset->data[sampleIndex].begin(), reverseDataset->data[sampleIndex].end(), reverseYrange);
		reverseCurve->setRawSamples(reverseXrange, reverseYrange, reverseDataset->nDimensions);
		reversePlot->replot();
	}

}

void DatasetViewer::on_pushButton_toggled(bool checked)
{
	showReverse = checked;
	prevIndex = -1;
	ui->plot->setVisible(!showReverse);
	ui->reversePlot->setVisible(showReverse);
	setSample(ui->sampleNumberSlider->value());
}
