#include <QApplication>
#include <QDesktopWidget>
#include <QStyleFactory>

#include "mainwindow.h"
#include <stdio.h>
#include <sstream>

#include "data/inputdataset.h"
#include "data/dataloader.h"
#include "data/dataexporter.h"

#include "reduction/reductionservice.h"

#include "clusterquality.h"

#include "common.h"

#include "tbb/task_scheduler_init.h"

#define L(a) std::cout << (a) << std::endl
#define L2(a,b) std::cout << (a) << (b) << std::endl
#define E(a) L(a); exit(1)


#ifdef PROFILEKM
#include "google/profiler.h"
#endif

#ifdef PROFILE_ALGO
#include "google/profiler.h"
#endif

void printHelp() {
	// List options, and reduction indexes
	L("Usage: ./paraspikes [options]");
	L("No options ==> GUI");
	L("");
	L("-f [filename]  File to load");
	L("-g [filename]  Cutfile");
	L("-r [reducIdx]  Which reduction to use (see trailing list)");
	L("-c [clustAlgo] Which clustering algo to use");
	L("   SPC (s)");
	L("    -[ float   Left limit");
	L("    -] float   Right limit");
	L("    -_ float   Step");
	L("   K means (k)");
	L("    -n int     Centroids");
	L("    -P int     Cutoff");
	L("-a             ASCII input (default: Axona)");
	L("-b             Run benchmark suite");
	L("  -I           Num iterations");
	L("-b             Run benchmark suite");
	L("-S             Srand(0)");
	L("-C             Num cores enabled");
	L("-h             You just did ");	
	L("");
	L("Reductions");

	QStringList reductions = ReductionService::availableReductions();

	int i=0;
	for(QStringList::iterator it = reductions.begin(); it != reductions.constEnd(); it++) {
		std::cout << i << ": " << it->toLatin1().data() << std::endl;
		i++;
	}
}

void runKmeansBenchmark(unsigned int numIterations, int cutoff, char *filename, int fileTypeIndex) {
	//TODO: also the 18 set..
	std::cout << "Running the kmeans benchmark suite with " << numIterations << " iterations." << std::endl;

	if(filename == NULL) {
		std::cerr << "No filename specified." << std::endl;
		exit(-1);
	}

	InputDataset *inputData = DataLoader::loadFile(filename, fileTypeIndex);
	if(inputData == NULL) {
		E("Could not load file");
	}

	// Reduce
	for(int reductionIndex = 0; reductionIndex < 1; ++reductionIndex) {
		std::cout << "Algo " << reductionIndex << std::endl;
		Reduction *reduction = ReductionService::getReductionIndex(reductionIndex);
		InputDataset *reducedDataset = reduction->reduce(inputData);

		L("Algorithm: K means");
		for(int k = 6; k < 7; ++k) {
			L2("Centroids: ", k);
			double globalSil = 0.0;
			uint64_t lowestNumTicks = UINT64_MAX;
#ifdef PROFILEKM
			ProfilerStart("profile/kmeans");
#endif
			for(int iterationIndex = 0; iterationIndex < numIterations; ++iterationIndex) {
				std::cout << "\tIteration " << iterationIndex << " of " << numIterations << "..." << std::endl;
				KMeans *km = new KMeans(reducedDataset->data, k, cutoff);
				//tbb::task_scheduler_init init(threads);
				uint64_t tempNumTicks = getticks();
				while(!km->convergenceStatus) {
					km->loopSerial();//loopParallel();
				}
#ifdef PROFILEKM
				ProfilerStop();
#endif
				tempNumTicks = getticks() - tempNumTicks;

				if(tempNumTicks < lowestNumTicks)
					lowestNumTicks = tempNumTicks;

				//km->convertToVectorFormat();

				//ClusteringAlgorithm *algo = km;
				//LratioQuality clusterQuality(algo->points, algo->lastComputedClusters);
				//LratioQuality clusterQuality(inputData->data, algo->lastComputedClusters);

				//clusterQuality.calculateLratioQualityAndIsolationDistance();
				//globalSil += clusterQuality.calculateAverageSilhouettes();
			}
			std::cout << "Minimum ticks taken: " << lowestNumTicks << std::endl;
			globalSil /= numIterations;
			std::cout << "Global silhouette: " << globalSil << std::endl;
		}

	}
}

void runSpcBenchmark() {
	//TODO: also the 18 set..
	for(int fileIndex = 1; fileIndex < 9; ++fileIndex) {
		std::stringstream filename;
		filename << "input/albert1/06041101." << fileIndex;
		std::cout << "Dataset " << filename.str() << std::endl;
		InputDataset *inputData = DataLoader::loadFile(filename.str().c_str(), 1);
		if(inputData == NULL) {
			E("Could not load file");
		}

		// Reduce
		for(int reductionIndex = 0; reductionIndex < 10; ++reductionIndex) {
			std::cout << "Algo " << reductionIndex << std::endl;
			Reduction *reduction = ReductionService::getReductionIndex(reductionIndex);
			InputDataset *reducedDataset = reduction->reduce(inputData);

			L("Algorithm: SPC");

			Spc *spc = new Spc(reducedDataset->data);

			// TODO for which temperature?
			spc->calculateSusceptibility(spc->initialGuessTemperature, true, NULL, false);

			ClusterQuality clusterQuality(spc->points, spc->lastComputedClusters);
			// LratioQuality clusterQuality(inputData->data, cutFileClusters);

			std::cout << "Silhouette: " << clusterQuality.calculateAverageSilhouettes() << std::endl;

		}
	}
}

void runBenchmark(char c, int numIterations, int cutoff, char *filename, int fileTypeIndex) {
	switch(c) {
	case 'k': runKmeansBenchmark(numIterations, cutoff, filename, fileTypeIndex);
		return;
	case 's': runSpcBenchmark();
		return;
	}
}

int missing(const char* what) {
	std::cerr << "Missing parameter: " << what << std::endl;
	return 1;
}

int cliMain(int argc, char *argv[]) {
	tbb::task_scheduler_init *init;
	int c;
	char *filename = NULL;
	int filetypeIndex = 1; // AXONA default
	char clusteringAlgorithm = '\0';
	char *cutFilename = NULL;
	int reductionIndex = 0;

	// SPC
	double leftLimit = 0.0;
	double rightLimit = 1.0;
	double step = 0.1;

	// kmeans
	int kmeansCentroids = 3;	

	int numThreads = -1;
	int numIterations = 1;
	int cutoff = 100;

	bool benchmark = false;
	char benchmarkType;

	srand(time(0));

	while ( (c = getopt(argc, argv, "SI:C:f:g:ar:c:o:h[:]:_:n:b:P:")) != -1) {
		switch(c) {
		case 'f': // Filename
			filename = optarg;
			break;

		case 'g': // Cut file
			cutFilename = optarg;
			break;

		case 'C':
			numThreads = atoi(optarg);
			break;

		case 'I':
			numIterations = atoi(optarg);
			break;

		case 'P':
			cutoff = atoi(optarg);
			break;

		case 'a': // Ascii input (Axona is default)
			filetypeIndex = 0;
			break;

		case 'r': // Reduction
			reductionIndex = atoi(optarg);
			break;

		case 'S': //Srand(0)
			srand(0);
			break;

		case 'c': // Clustering algorithm
			clusteringAlgorithm = optarg[0];
			break;

		case 'n': // N centroids
			kmeansCentroids = atoi(optarg);
			break;

		case '[': // Left limit
			leftLimit = atof(optarg);
			break;

		case ']': // Right limit
			rightLimit = atof(optarg);
			break;

		case '_': // Step
			step = atof(optarg);
			break;

		case 'h': // Help
			printHelp();
			return 0;

		case 'b':
			benchmark = true;
			benchmarkType = optarg[0];
			break;
		}
	}

	std::cout << "TBB: Default number of threads: " << tbb::task_scheduler_init::default_num_threads() << std::endl;
	if(numThreads > 0) {
		std::cout << "TBB: - Manually adjusted numThreads to " << numThreads << std::endl;
			init = new tbb::task_scheduler_init(numThreads);
	}

	if(benchmark) {
		runBenchmark(benchmarkType, numIterations, cutoff, filename, filetypeIndex);
		return 0;
	}

	if(filename == NULL) {
		return missing("Filename");
	}

	if(clusteringAlgorithm == '\0') {
		return missing("Clustering algorithm");
	}

	std::cout << "File: " << filename << std::endl;
	std::cout << "Filetype: " << ((filetypeIndex==0)?"ASCII":"Axona") << std::endl;
	std::cout << "Reduction: " << ReductionService::availableReductions()[reductionIndex].toLatin1().data() << std::endl;

	// Then run everything!

	InputDataset *inputData = DataLoader::loadFile(filename, filetypeIndex);
	if(inputData == NULL) {
		E("Could not load file");
	}

	// Reduce
	tbb::tick_count reductionStart = tbb::tick_count::now();
	Reduction *reduction = ReductionService::getReductionIndex(reductionIndex);
	InputDataset *reducedDataset = reduction->reduce(inputData);
	tbb::tick_count reductionEnd = tbb::tick_count::now();
	std::cout << "time for reduction: " << (reductionEnd - reductionStart).seconds() << std::endl;

	// Expected output for algo:
	ClusteringAlgorithm *algo = NULL;


	tbb::tick_count algoStart = tbb::tick_count::now();
#ifdef PROFILE_ALGO
	ProfilerStart("profile/algo");
#endif

	// Run algorithm
	switch(clusteringAlgorithm) {
	case 's': // SPC
	{
		L("Algorithm: SPC");
		L2("-Left limit: ", leftLimit);
		L2("-Right limit: ", rightLimit);
		L2("-Step: ", step);

		Spc *spc = new Spc(reducedDataset->data);
		spc->plotSusceptibility(leftLimit, rightLimit, step);

		algo = spc;
		// TODO some hint on time remaining would be nice
	}
		break;
	case 'k':
	{
		L("Algorithm: K means");
		L2("-Centroids: ", kmeansCentroids);
		KMeans *km = new KMeans(reducedDataset->data, kmeansCentroids, cutoff);
		while(km->convergenceStatus != 2) {
			km->OnLoop();
		}		
		km->convertToVectorFormat();

		//PcaPerChannelReduction pcapc;
		//InputDataset *pcaPerChannel = pcapc.reduce(inputData);
		//LratioQuality foo(pcaPerChannel->data, km.lastComputedClusters);
		//foo.calculateClusterQuality();
		algo = km;
	}
		break;
	default:
		std::cerr << "Unknown algorithm: " << clusteringAlgorithm << std::endl;
		return 1;
	}

#ifdef PROFILE_ALGO
	ProfilerStop();
#endif

	tbb::tick_count algoEnd = tbb::tick_count::now();
	std::cout << "time for algo total: " << (algoEnd - algoStart).seconds() << std::endl;

	// Done clustering, check results
//	L("Quality.....");

//	//remove empty clusters. TODO: move elsewhere probably
//	std::vector<std::vector<int> >::iterator it = algo->lastComputedClusters.begin();
//	while(it != algo->lastComputedClusters.end()) {
//		if((*it).empty())
//			it = algo->lastComputedClusters.erase(it);
//		else
//			++it;
//	}

//	/*{
//		// Saving without zero-cluster
//		std::vector<std::vector<int> > cutFileClusters;
//		DataLoader::loadCutFile(cutFilename, algo->points, algo->lastComputedClusters, cutFileClusters);
//		L("Saving files witout cluster 0...");
//		char withoutZeroFilename[200] = {0};
//		sprintf(withoutZeroFilename, "%s.nonoise", filename);
//		DataExporter e;
//		e.saveWithoutClusterZero(inputData, cutFileClusters, withoutZeroFilename);
//		return 0;
//	}*/

//	ClusterQuality clusterQuality(algo->points, algo->lastComputedClusters);
//	// LratioQuality clusterQuality(inputData->data, cutFileClusters);
//	//clusterQuality.calculateLratioQualityAndIsolationDistance();
//	clusterQuality.calculateAverageSilhouettes();

//	if(cutFilename != NULL) {
//		// Measure cut file data
//		L("Comparing to cut file...");
//		std::vector<std::vector<int> > cutFileClusters;
//		DataLoader::loadCutFile(cutFilename, algo->points, algo->lastComputedClusters, cutFileClusters);

//		std::cout << "Clusters: " << std::endl;
//		for(int i = 0; i < cutFileClusters.size(); ++i)
//			std::cout << "\t#" << i << ": " << cutFileClusters[i].size() << std::endl;


//		L("Checking Fmeasure for cutfile against itself");
//		ClusterQuality cutCutQuality(algo->points, algo->lastComputedClusters);
//		//cutCutQuality.calculateAverageSilhouettes();
//		//return 0;
//		cutCutQuality.printFmeasurePrecisionAndRecall(cutFileClusters);




//	} else {
//		std::cout << "No cutfile specified -- no comparisons performed" << std::endl;

//		std::cout << "Clusters: " << std::endl;
//		for(int i = 0; i < algo->lastComputedClusters.size(); ++i) {
//			std::cout << "\t#" << i << ": " << algo->lastComputedClusters[i].size() << std::endl;
//			/*std::cout << "\t\t";
//			for(int j = 0; j < algo->lastComputedClusters[i].size(); ++j) {
//				int index = algo->lastComputedClusters[i][j];
//				std::cout << "[" << algo->points[index][0] << ", " << algo->lastComputedClusters[index][1] << "] ";
//			}
//			std::cout << std::endl;*/
//		}
//	}


	L("Done");

	return 0;
}

int main(int argc, char *argv[])
{	
	std::cout << "Paraspikes, built: " << __DATE__ << " " << __TIME__ << std::endl << "==========" << std::endl;
	if(argc ==1) {
		QApplication app(argc, argv);

		//app.setStyle("cleanlooks");
		MainWindow window;
		window.resize(1280, 960);
		int desktopArea = QApplication::desktop()->width() *
						 QApplication::desktop()->height();
		int widgetArea = window.width() * window.height();
		if (((float)widgetArea / (float)desktopArea) < 0.75f)
			window.show();
		else
			window.showMaximized();
		return app.exec();
	} else {
		return cliMain(argc, argv);
	}
}
