#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "data/dataloader.h"
#include <QMessageBox>
#include <QVector>
#include <QFileDialog>
#include <QLayout>
#include "clusterquality.h"
#include <macros.h>

#include "reduction/reductionservice.h"
#include "data/dataexporter.h"
#include "filtering/sdfilter.h"

#include <iostream>
#include <fstream>

#include <qwt_symbol.h>


MainWindow::MainWindow(QWidget *parent) :
	QMainWindow(parent),
	currentClusteringIndex(CLUSTER_NONE),
	ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    ui->fileChooser->addFormats(DataLoader::filetypes());
	ui->fileChooser->setFilename("input/180501.1");

	ui->reductionChooser->addSchemes(ReductionService::availableReductions());

	ui->datasetInformation->setTitle("Input dataset information");
	ui->reducedDatasetInformation->setTitle("Reduced dataset information");
	connect(ui->datasetInformation, SIGNAL(viewDataset(InputDataset*)), this, SLOT(viewDataset(InputDataset*)));
	connect(ui->reducedDatasetInformation, SIGNAL(viewDataset(InputDataset*)), this, SLOT(viewDataset(InputDataset*)));

	connect(ui->susceptibilityPlot, SIGNAL(temperatureChanged(double)), ui->clusterSizePlot, SLOT(setTemperature(double)));
	connect(ui->clusterSizePlot, SIGNAL(temperatureChanged(double)), ui->susceptibilityPlot, SLOT(setTemperature(double)));

	ui->spikeResults->setSpacing(10);
//	ui->plotSpikeClustersButton->hide();
//	ui->resultsTab->setEnabled(false);

	//default to spc, hide kmeans stuff. todo: should just hide a container, and in designer since that is where spc is defaulted..
	ui->centroidsInput->setVisible(false);
	ui->centroidsLabel->setVisible(false);

    inputData = NULL;
    reducedInputData = NULL;
	spc = NULL;
	km = NULL;
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void MainWindow::on_fileChooser_openFile(QString filename, int index)
{
	if(!QFile::exists(filename)) {
		QMessageBox::warning(this, "File not found", "The chosen file could not be loaded.");
		return;
	}
    //TODO progressbar + async?

    //TODO delete old data
    if(inputData != NULL) {
        delete inputData;
        inputData = NULL;
    }

	currentClusteringIndex = CLUSTER_NONE;

	inputData = DataLoader::loadFile(filename, index);

	if(inputData == NULL) {
		QMessageBox::warning(this, "Could not read file", "The file could not be read. It could be in another format than specified");
		return;
	}

	ui->datasetInformation->setDataset(inputData);


	ui->reductionChooser->clickReduceButton(); //TODO or could be just manual

//	ui->resultsTab->show();
	//ui->plotSpikeClustersButton->show();
	//ui->tabWidget->setTabEnabled(ui->tabWidget->currentIndex(), true);
//	ui->resultsTab->setEnabled(true);
}

void MainWindow::on_actionAbout_triggered()
{
	QMessageBox::information(this, "About ParaSpikes", "Developed by\nThomas Bergheim and Arve Skogvold\n\nMaster thesis\nNorwegian University of Science and Technology\n2011");
}

void MainWindow::on_plotSusceptibilityButton_clicked()
{
	if(spc == NULL) {
        QMessageBox::warning(this, "Error", "You have to load a dataset before plotting susceptibility.");
        return;
    }

    double minTemperature = ui->minTempSpinner->value();
    double maxTemperature = ui->maxTempSpinner->value();
    double step = ui->stepSpinner->value();

	spc->finishness = 0;

	susceptibilityThread = new SusceptibilityThread(this, spc, minTemperature, maxTemperature, step);
	connect(susceptibilityThread, SIGNAL(finished()), this, SLOT(susceptibilityThreadFinished()));
	susceptibilityThread->start();

	//progressThread = new ProgressCheckingThread(this, spc);
	progressTimer = new QTimer(this);
	progressTimer->setInterval(500);
	progressTimer->setSingleShot(false);
	connect(progressTimer, SIGNAL(timeout()), this, SLOT(updateProgress()));

	progressDialog = new QProgressDialog(this, Qt::Dialog);
	connect(progressDialog, SIGNAL(canceled()), this, SLOT(cancelSusceptibilityPlot()));
	progressDialog->setWindowTitle("Plotting susceptibility...");
	progressDialog->setModal(true);
	progressDialog->setLabelText("Estimating remaining time...");
	progressDialog->setValue(0);
	progressDialog->show();

	calculateStart = new QTime();
	calculateStart->start();

	progressTimer->start();


}

void MainWindow::susceptibilityThreadFinished() {
	disconnect(progressTimer, SIGNAL(timeout()), this, SLOT(updateProgress()));
	progressTimer->stop();


	ui->susceptibilityPlot->setSuPlot(*spc->susceptibilityPlot);

	ui->clusterSizePlot->setData(*spc->correlationClusterSizesPlot);

	delete progressTimer;
	progressTimer = NULL;
	delete progressDialog;
	delete susceptibilityThread;
	delete calculateStart;
	setEnabled(true);
}

void MainWindow::on_susceptibilityPlot_temperatureChanged(double newTemperature)
{
	if(spc != NULL) {
		spc->calculateSusceptibility(newTemperature, true, NULL, true);

		ClusterQuality cq(spc->points, spc->lastComputedClusters);

		cq.calculateLratioQualityAndIsolationDistance();
		cq.calculateAverageSilhouettes();
		cq.printFmeasurePrecisionAndRecall(spc->cutFileIds);
	}
    //TODO display the result somehow
}

void MainWindow::updateProgress() {
	double frac = spc->finishness;
	if(frac <= 0)
		return;

	double remaining;
	if(frac == lastFrac) {
		remaining = lastRemaining - 0.5;
	} else {
		progressDialog->setValue((int)(frac*100));

		remaining = (1-frac) * 0.001 * calculateStart->elapsed()/frac;

	}

	QTime t;
	t = t.addSecs((int)remaining);
	progressDialog->setLabelText(QString("Time remaining: %1").arg(t.toString("h:mm:ss")));

	lastFrac = frac;
	lastRemaining = remaining;
}

ClusteringAlgorithm *MainWindow::getCurrentClusteringAlgorithm()
{
	ClusteringAlgorithm *ca = NULL;

	std::cout << "ok: " << currentClusteringIndex << std::endl;

	switch(currentClusteringIndex)
	{
	case SPC:
		assert(spc != NULL);
		ca = spc;
		break;
	case KMEANS:
		assert(km != NULL);
		ca = km;
		break;
	default:
		break;
	}

	return ca;
}

void MainWindow::on_reductionChooser_performReduction(int index)
{
	if(reducedInputData != NULL) {
		delete reducedInputData;
		reducedInputData = NULL;
	}
	ui->reducedDatasetInformation->setDataset(reducedInputData);

	if(inputData == NULL) {
		QMessageBox::information(this, "Reduction problems", "Could not reduce non-existing dataset");
		return;
	}
	Reduction *r = ReductionService::getReductionIndex(index);
	reducedInputData = r->reduce(inputData);
	ui->reducedDatasetInformation->setDataset(reducedInputData);

	//ui->clusterButton->click();
}

DatasetViewer* MainWindow::viewDataset(InputDataset *dataset) {
	DatasetViewer *viewer = new DatasetViewer(dataset,this);
	viewer->setModal(true);
	viewer->show();
	return viewer;
}

void MainWindow::cancelSusceptibilityPlot() {
	setEnabled(false);
	progressTimer->stop();
	spc->cancelSusceptibilityPlot();
}

void MainWindow::on_plotSpikeClustersButton_clicked()
{
	if(inputData == NULL)
	{
		QMessageBox::information(this, "Error", "You can't plot an unloaded data set.");
		return;
	}


	InputDataset *datasetToBePlotted;
	if(ui->plotReducedDatasetToggle->isChecked()) {
		std::cerr << "Plotting reduced" << std::endl;
		if(ui->plotReverseReducedToggle->isChecked()) {
			datasetToBePlotted = reducedInputData->reduceClass->reverse(reducedInputData); // TODO memory leak
		} else {
			datasetToBePlotted = reducedInputData;
		}
	} else {
		std::cerr << "Plotting original" << std::endl;
		datasetToBePlotted = inputData;
	}

	static std::vector <QwtPlot *> currentSpikePlotList;

	ClusteringAlgorithm *ca = getCurrentClusteringAlgorithm();
	double *xdim = new double[datasetToBePlotted->nDimensions];
	double *ydim = new double[datasetToBePlotted->nDimensions];

	for(unsigned int i = 0; i < datasetToBePlotted->nDimensions; ++i) {
		xdim[i] = i;
	}

	for(std::vector <QwtPlot *>::iterator it = currentSpikePlotList.begin(); it != currentSpikePlotList.end(); ++it)
	{
		delete *it;
	}
	currentSpikePlotList.clear();

	if(ca == NULL)
	{
		QMessageBox::information(this, "Missing clusters", "You have to perform clustering first.");
		//QwtPlot *spikes = new QwtPlot(ui->resultsTab);
		//spikes->setFixedSize(400, 200);
		//ui->spikeResults->addWidget(spikes);
	}
	else
	{
		std::cout << "Drawing " << ca->lastComputedClusters.size() << " clusters." << std::endl;

		SdFilter sdFilter(datasetToBePlotted); // TODO should we identify noise in datasetToBePlotted, or always in input dataset?
		int howMany = max(ca->lastComputedClusters.size(), ca->cutFileIds.size());
		for(int cluster = 0; cluster <  howMany; ++cluster) {
			if(ca->lastComputedClusters[cluster].size() == 0)
				continue;

			// Standard deviation filtering
			std::vector<int> *filteredSamples = sdFilter.findNoiseSpikeIndexes(ca->lastComputedClusters[cluster], ui->sdNoiseRemovalSpinBox->value());



			std::vector <int> *clusterToBePlotted = NULL;
			if(ui->filterOutsideSdChannelCheckbox->isChecked() && ui->sdNoiseRemovalSpinBox->value() > 0) {
				clusterToBePlotted = filteredSamples;
			} else {
				clusterToBePlotted = &ca->lastComputedClusters[cluster];
			}

			QwtPlot *spikes = new QwtPlot(ui->resultsTab);
			//spikes->setFixedSize(400, 200);
			spikes->setCanvasBackground(QColor(Qt::white));


			// Plot/color standard deviation channel
			if(ui->plotSdChannelCheckbox->isChecked()) {
				QwtPlotCurve *topStdCurve = new QwtPlotCurve();
				topStdCurve->setRenderHint(QwtPlotItem::RenderAntialiased);
				topStdCurve->setBaseline(-1000);
				topStdCurve->setBrush(QBrush(Qt::gray));
				for(unsigned int i=0; i < sdFilter.meanSpike.size(); i++) {
					ydim[i] = sdFilter.meanSpike[i] + ui->sdNoiseRemovalSpinBox->value() * sdFilter.stdDev[i];
				}
				topStdCurve->setSamples(xdim, ydim, sdFilter.stdDev.size());
				topStdCurve->attach(spikes);

				QwtPlotCurve *bottomStdCurve = new QwtPlotCurve();
				bottomStdCurve->setRenderHint(QwtPlotItem::RenderAntialiased);
				bottomStdCurve->setBaseline(-1000);
				bottomStdCurve->setBrush(QBrush(spikes->canvasBackground()));
				for(unsigned int i=0; i < sdFilter.meanSpike.size(); i++) {
					ydim[i] = sdFilter.meanSpike[i] - ui->sdNoiseRemovalSpinBox->value() * sdFilter.stdDev[i];
				}
				bottomStdCurve->setSamples(xdim, ydim, sdFilter.stdDev.size());
				bottomStdCurve->attach(spikes);
			}





			// Plot the spikes
			if(ui->plotSpikesCheckbox->isChecked()) {
				for(unsigned int spikeIndexInCluster = 0; spikeIndexInCluster < clusterToBePlotted->size(); ++spikeIndexInCluster) {
					int spikeIndex = (*clusterToBePlotted)[spikeIndexInCluster];

					QwtPlotCurve *curve = new QwtPlotCurve();
					curve->setRenderHint(QwtPlotItem::RenderAntialiased);
					curve->setPen(QPen(QColor::fromHsv(255*spikeIndexInCluster/clusterToBePlotted->size(),255,255)));

					std::copy(datasetToBePlotted->data[spikeIndex].begin(), datasetToBePlotted->data[spikeIndex].end(), ydim);
					curve->setSamples(xdim, ydim, datasetToBePlotted->nDimensions);
					curve->attach(spikes);
				}
			}

			// Add channel separators
			int samplesPerChannel = datasetToBePlotted->nSamples / datasetToBePlotted->nChannels;
			if(datasetToBePlotted->nDimensions == 200 || datasetToBePlotted->nDimensions == 256) {
				samplesPerChannel = 50;
			}
			for(unsigned int i=1;i < datasetToBePlotted->nChannels ;i++) {
				QwtPlotMarker *mrk = new QwtPlotMarker();
				mrk->setLineStyle(QwtPlotMarker::VLine);
				mrk->setLinePen(QPen(Qt::red, 0, Qt::DashLine));
				mrk->setXValue(i*samplesPerChannel - 0.5);
				mrk->attach(spikes);
			}

			// Set scale
			if(datasetToBePlotted->nDimensions == 200 || datasetToBePlotted->nDimensions == 256) {
				spikes->setAxisScale(QwtPlot::yLeft, -150, 150);
			}

			// Plot mean
			if(ui->plotMeansCheckbox->isChecked()) {
				QwtPlotCurve *meanCurve = new QwtPlotCurve();
				meanCurve->setRenderHint(QwtPlotItem::RenderAntialiased);
				meanCurve->setPen(QPen(QBrush(QColor(Qt::red)),3, Qt::DotLine));
				std::copy(sdFilter.meanSpike.begin(), sdFilter.meanSpike.end(), ydim);
				meanCurve->setSamples(xdim, ydim, sdFilter.meanSpike.size());
				meanCurve->attach(spikes);
			}


			// Plot cutfile
			if(!ca->cutFileIds.empty() && cluster < ca->cutFileIds.size() ) {
				QwtPlotCurve *meanCurve = new QwtPlotCurve();
				meanCurve->setRenderHint(QwtPlotItem::RenderAntialiased);
				meanCurve->setPen(QPen(QBrush(QColor(Qt::green)),3, Qt::DotLine));
				std::vector<double> meanSpike(datasetToBePlotted->nDimensions, 0);

				// Calculate mean
				for(std::vector<int>::const_iterator it = ca->cutFileIds[cluster].begin(); it != ca->cutFileIds[cluster].end(); ++it) {
					int spikeIndex = *it;

					for(unsigned int dimIndex = 0; dimIndex < datasetToBePlotted->nDimensions; dimIndex++) {
						meanSpike[dimIndex] += datasetToBePlotted->data[spikeIndex][dimIndex];
					}
				}

				for(unsigned int dimIndex = 0; dimIndex < meanSpike.size(); dimIndex++) {
					meanSpike[dimIndex] /= ca->cutFileIds[cluster].size();
				}

				std::copy(meanSpike.begin(), meanSpike.end(), ydim);


				meanCurve->setSamples(xdim, ydim, meanSpike.size());
				meanCurve->attach(spikes);
			}


			ui->spikeResults->addWidget(spikes);

			currentSpikePlotList.push_back(spikes);
		}

		delete[] xdim;
		delete[] ydim;
	}
}

void MainWindow::on_clusterButton_clicked()
{
	//TODO: make more robust
	currentClusteringIndex = ui->clusteringTypeComboBox->currentIndex();

	//TODO: cleean code!
	switch(currentClusteringIndex)
	{
	case SPC:
		if(spc != NULL) {
			delete spc;
		}
		spc = new Spc(reducedInputData->data);
		ui->suggestionLabel->setText(QString::number(spc->initialGuessTemperature));
		break;
	case KMEANS:
		if(km != NULL) {
			delete km;
		}
		km = new KMeans(reducedInputData->data, ui->centroidsInput->text().toInt(), 100);
		while(km->convergenceStatus != 2) {
			km->OnLoop();
		}
		km->convertToVectorFormat();

		ui->plotSpikeClustersButton->click();
		break;
	default:
		break;
	}


}

void MainWindow::on_clusteringTypeComboBox_currentIndexChanged(QString selected)
{
	if(selected == "k-means") {
		ui->centroidsLabel->setVisible(true);
		ui->centroidsInput->setVisible(true);
	} else {
		ui->centroidsInput->setVisible(false);
		ui->centroidsLabel->setVisible(false);
	}
}

void MainWindow::on_compareButton_clicked()
{
	if(inputData == NULL || reducedInputData == NULL) {
		QMessageBox::warning(this, "Cannot plot", "You have to load a file first.");
		return;
	}
	DatasetViewer *v1 = viewDataset(inputData);
	DatasetViewer *v2 = viewDataset(reducedInputData);

	connect(v1, SIGNAL(indexChanged(int)), v2, SLOT(setSample(int)));
	connect(v2, SIGNAL(indexChanged(int)), v1, SLOT(setSample(int)));

}

void MainWindow::on_plotScatterButton_clicked()
{
	if(reducedInputData == NULL) {
		QMessageBox::warning(this, "Cannot plot", "You have to load a file first.");
		return;
	}

	ClusteringAlgorithm *c = getCurrentClusteringAlgorithm();

	delete ui->scatterPlot;
	ui->scatterPlot = new QwtPlot(ui->scatterPlotTab);
	ui->scatterPlotContainer->addWidget(ui->scatterPlot);

	ui->scatterPlot->setCanvasBackground(QColor(Qt::white));
	//TODO: add clear scatter

	// no clustering performed

	InputDataset *scatterData;
	if(ui->scatterPcaCheckbox->isChecked()) {
		PcaReduction red;
		scatterData = red.reduce(reducedInputData, 2);
	} else {
		scatterData = reducedInputData;
	}

	if(!c) {
		if(scatterData->nDimensions < 2)
			return;

		QwtPlotCurve *curve = new QwtPlotCurve();
		QPen pen;
		//pen.setColor(QColor::fromHsv(255*cluster/clustering->size(),255,255));
		pen.setWidth(2);
		curve->setPen(Qt::NoPen);
		QwtSymbol *s = new QwtSymbol(QwtSymbol::XCross, Qt::NoBrush, pen, QSize(5,5));
		curve->setSymbol(s);
		curve->setStyle(QwtPlotCurve::Dots);

		double *x = new double[scatterData->nSamples];
		double *y = new double[scatterData->nSamples];

		for(unsigned int pointIndex = 0; pointIndex < scatterData->nSamples; ++pointIndex) {
			x[pointIndex] = scatterData->data[pointIndex][0];
			y[pointIndex] = scatterData->data[pointIndex][1];
		}

		curve->setRawSamples(x, y, scatterData->nSamples);

		curve->attach(ui->scatterPlot);

		//delete[] x;
		//delete[] y;

	} else {

		if(c->nDimensions < 2)
			return;

		// Plot cutfile?
		std::vector<std::vector<int> > *clustering;
		if(!ui->cutFileScatterCheckbox->isChecked()) {
			clustering = &c->lastComputedClusters;
		} else {
			if(c->cutFileIds.size() == 0) {
				QMessageBox::warning(this, "No cutfile loaded", "You need to load a cutfile before plotting it");
				return;
			}
			clustering = &c->cutFileIds;
		}

		unsigned int minClusterSize = ui->minClusterSizeSpinBox->value();


		// Plot the unclustered
		if(ui->includeUnclusteredCheckbox->isChecked())
		{
			QwtPlotCurve *curve = new QwtPlotCurve();
			QPen pen;
			//pen.setColor(QColor::fromHsv(255*cluster/clustering->size(),255,255));
			pen.setWidth(2);
			curve->setPen(Qt::NoPen);
			QwtSymbol *s = new QwtSymbol(QwtSymbol::XCross, Qt::NoBrush, pen, QSize(5,5));
			curve->setSymbol(s);
			curve->setStyle(QwtPlotCurve::Dots);

			double *x = new double[scatterData->nSamples];
			double *y = new double[scatterData->nSamples];

			for(unsigned int pointIndex = 0; pointIndex < scatterData->nSamples; ++pointIndex) {
				x[pointIndex] = scatterData->data[pointIndex][0];
				y[pointIndex] = scatterData->data[pointIndex][1];
			}

			curve->setRawSamples(x, y, scatterData->nSamples);

			curve->attach(ui->scatterPlot);
		}


		for(unsigned int cluster = 0; cluster < clustering->size(); ++cluster) {
			if(clustering->size() < minClusterSize)
				continue;

			if(cluster==0 && !ui->includeCluster0Checkbox->isChecked())
				continue;

			QwtPlotCurve *curve = new QwtPlotCurve();
			QPen pen;
			pen.setColor(QColor::fromHsv(255*cluster/clustering->size(),255,255));
			pen.setWidth(2);
			curve->setPen(Qt::NoPen);
			QwtSymbol *s = new QwtSymbol(QwtSymbol::Style(cluster % QwtSymbol::Hexagon), Qt::NoBrush, pen, QSize(5,5));
			curve->setSymbol(s);

			curve->setStyle(QwtPlotCurve::Dots);

			int currentClusterSize = (*clustering)[cluster].size();

			double *x = new double[currentClusterSize];
			double *y = new double[currentClusterSize];

			for(int pointIndex = 0; pointIndex < currentClusterSize; ++pointIndex) {
				x[pointIndex] = scatterData->data[(*clustering)[cluster][pointIndex]][0];
				y[pointIndex] = scatterData->data[(*clustering)[cluster][pointIndex]][1];
			}

			curve->setRawSamples(x, y, currentClusterSize);

			curve->attach(ui->scatterPlot);

			//delete[] x;
			//delete[] y;
		}

	}
	ui->scatterPlot->replot();

}

void MainWindow::on_saveCutFileButton_clicked()
{
	QString filename = QFileDialog::getSaveFileName(this, "Save cutfile as", "output/", "Cut file (*.cut)");

	if(!filename.size())
		return;

	DataExporter de;
	ClusteringAlgorithm *c = getCurrentClusteringAlgorithm();

	de.saveCutFile(c->clusterIds, filename.toAscii());


}

void MainWindow::on_loadCutFileBtn_clicked()
{
	QString filename = QFileDialog::getOpenFileName(this, "Open cutfile", "input/", "Cut file (*.cut)");

	if(!filename.size())
		return;

	ClusteringAlgorithm *c = getCurrentClusteringAlgorithm();
	DataLoader::loadCutFile(filename.toStdString().c_str(), c->points, c->lastComputedClusters, c->cutFileIds);
}

void MainWindow::on_useReducedAsInput_clicked()
{
		ui->reductionChooser->setIndex(0);
		inputData = new InputDataset(*reducedInputData);
		ui->datasetInformation->setDataset(inputData);
}

void MainWindow::on_silhouetteButton_clicked()
{
	ClusteringAlgorithm *ca = getCurrentClusteringAlgorithm();
	if(ca == NULL) {
		QMessageBox::warning(this, "No clustering results available", "You need to perform clustering first");
		return;
	}
	ClusterQuality lrc(ca->points, ca->lastComputedClusters);
	lrc.calculateAverageSilhouettes(); // TODO export for display in matlab?
}

void MainWindow::on_lRatioButton_clicked()
{
	ClusteringAlgorithm *ca = getCurrentClusteringAlgorithm();
	if(ca == NULL) {
		QMessageBox::warning(this, "No clustering results available", "You need to perform clustering first");
		return;
	}
	std::cout << "L-ratio is calculated using Energy PCA on original dataset" << std::endl;
	EnergyPcaReduction red;
	InputDataset *d = red.reduce(inputData);
	ClusterQuality lrc(d->data, ca->lastComputedClusters);
	lrc.calculateLratioQualityAndIsolationDistance();

	delete d;
}

void MainWindow::on_fmeasureButton_clicked()
{
	ClusteringAlgorithm *ca = getCurrentClusteringAlgorithm();
	if(ca == NULL) {
		QMessageBox::warning(this, "No clustering results available", "You need to perform clustering first");
		return;
	}

	if(ca->cutFileIds.size() == 0) {
		QMessageBox::warning(this, "No cutfile loaded", "You need to load a cutfile before calculating F-measure");
		return;
	}

	ClusterQuality lrc(ca->points, ca->lastComputedClusters);
	lrc.printFmeasurePrecisionAndRecall(ca->cutFileIds);
}

void MainWindow::on_export3dButton_clicked()
{
	ClusteringAlgorithm *ca = getCurrentClusteringAlgorithm();
	if(ca == NULL) {
		QMessageBox::warning(this, "No clustering results available", "You need to perform clustering first");
		return;
	}

	std::vector<std::vector<int> > *clustering;
	if(!ui->cutFileScatterCheckbox->isChecked()) {
		clustering = &ca->lastComputedClusters;
	} else {
		if(ca->cutFileIds.size() == 0) {
			QMessageBox::warning(this, "No cutfile loaded", "You need to load a cutfile before plotting it");
			return;
		}
		clustering = &ca->cutFileIds;
	}

	PcaReduction pca;
	InputDataset *pcaReducedData = pca.reduce(reducedInputData, 3);

	int numClusters = clustering->size();

	std::ofstream fout;
	fout.open("clusters3d_out.py");
	fout << "clusters = []" << std::endl;

	for(int cluster=0; cluster < numClusters; cluster++) {
		fout << "clusters.append([" << std::endl;
		for(unsigned int spikeIndexInCluster=0; spikeIndexInCluster < (*clustering)[cluster].size(); spikeIndexInCluster++) {
			for(int dim=0;dim<3;dim++) {
				fout << pcaReducedData->data[ (*clustering)[cluster][spikeIndexInCluster] ][dim] << ", ";
			}
			fout << std::endl;
		}
		fout << "\n])" << std::endl;

	}

	fout.close();
	QMessageBox::information(this, "3d plot exported", "View the file by running ./plot3d.py");
}

void MainWindow::on_pretendCutFileButton_clicked()
{
	ClusteringAlgorithm *ca = getCurrentClusteringAlgorithm();
	if(ca != NULL && !ca->cutFileIds.empty()) {
		ca->lastComputedClusters = ca->cutFileIds;
	}
}
