#include "pcareduction.h"
#include <vector>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_eigen.h>
#include <string.h>
#include <math.h>
#include <gsl/gsl_blas.h>
#include <iostream>
#include <stdio.h>

#include "macros.h"

PcaReduction::PcaReduction()
{
}

static void
printMatrixTheHardWay(gsl_matrix*m)
{
	/* print matrix the hard way */
	printf("Matrix m\n");
	for (unsigned int row=0;row<m->size1;row++)
	{
		for (unsigned int col=0;col<m->size2;col++)
		{
			printf("%f ",gsl_matrix_get(m,row,col));
		}
		printf("\n");
	}
	std::cout << std::endl;
}

InputDataset *PcaReduction::reduce(const InputDataset *input) {
	return reduce(input, -1);
}

InputDataset *PcaReduction::reduce(const InputDataset *input, int numComponents)
{
	if(numComponents == -1) {
		numComponents = 2; // TODO dependent on data?
	}

	int numSpikes = input->data.size();
	int numFeatures = input->data[0].size();

	gsl_matrix *mdev = gsl_matrix_alloc(numFeatures, numSpikes);
	gsl_matrix *m = gsl_matrix_alloc(numFeatures, numSpikes);

	double *mean = new double[numFeatures];
	memset(mean, 0, sizeof(double)*numFeatures);

#pragma omp parallel for
	for(int i = 0; i < numSpikes; ++i) {
		for(int j = 0; j < numFeatures; ++j) {
			mean[j] += input->data[i][j];
		}
	}

#pragma omp parallel for
	for(int j = 0; j < numFeatures; ++j) {
		mean[j] /= numSpikes;
	}
	std::cout << std::endl;

	gsl_matrix *covarianceMatrix = gsl_matrix_alloc(numFeatures, numFeatures);
	gsl_vector_view a, b;

#pragma omp parallel for
	for(unsigned int spike = 0; spike < input->data.size(); ++spike) {
		for(int feature = 0; feature < numFeatures; ++feature) {
			gsl_matrix_set(m, feature, spike, input->data[spike][feature]);
			gsl_matrix_set(mdev, feature, spike, input->data[spike][feature] - mean[feature]);
		}
	}


	for (int dim1 = 0; dim1 < numFeatures; dim1++) {
#pragma omp parallel for
		for (int dim2 = dim1; dim2 < numFeatures; dim2++) {
			double v;
			a = gsl_matrix_row(mdev, dim1);
			b = gsl_matrix_row(mdev, dim2);
			//v = gsl_stats_covariance_m(a.vector.data, a.vector.stride, b.vector.data, b.vector.stride, a.vector.size, mean[dim1], mean[dim2]); //TODO wrong mean
			v = gsl_stats_covariance(a.vector.data, a.vector.stride, b.vector.data, b.vector.stride, a.vector.size);
			gsl_matrix_set(covarianceMatrix, dim1, dim2, v);
		}
	}

#pragma omp parallel for
	for (int dim1 = 1; dim1 < numFeatures; dim1++) {
		for (int dim2 = 0; dim2 < dim1; dim2++) {
			gsl_matrix_set(covarianceMatrix, dim1, dim2, gsl_matrix_get(covarianceMatrix, dim2, dim1));
		}
	}


//	WHEREAMI();
//	std::cout << "covariance" << std::endl;
//	printMatrixTheHardWay(covarianceMatrix);

	gsl_vector *eigenValues = gsl_vector_alloc (numFeatures);
	gsl_matrix *eigenVectors = gsl_matrix_alloc (numFeatures, numFeatures);

	gsl_eigen_symmv_workspace * work = gsl_eigen_symmv_alloc(numFeatures);
	gsl_eigen_symmv (covarianceMatrix, eigenValues, eigenVectors, work);

	gsl_eigen_symmv_free(work);

	gsl_eigen_symmv_sort (eigenValues, eigenVectors, GSL_EIGEN_SORT_ABS_DESC);

//	WHEREAMI();
//	gsl_vector_fprintf(stdout, eigenValues, "%f\n");
//	WHEREAMI();
//	printMatrixTheHardWay(eigenVectors);

//	std::vector<double> cumulativeEigenValues(numFeatures); // TODO used for Z scaling, re wikipedia

//	cumulativeEigenValues[0] = eigenValue->data[0];
//	for(int i=1; i < numFeatures; i++) {
//		cumulativeEigenValues[i] = cumulativeEigenValues[i-1] + eigenValue->data[i];
//	}

//	std::vector<double> stdDev(numFeatures); //TODO has to be done before, as the eigen functions destroy the matrix

//	for(int i=0;i<numFeatures;i++) {
//		stdDev[i] = sqrt(gsl_matrix_get(covarianceMatrix, i, i));
//	}

	gsl_matrix *subsetEigenVectorsMatrix = gsl_matrix_alloc(numFeatures, numComponents);

#pragma omp parallel for
	for(int i=0; i<numFeatures; i++) {
		for(int j=0; j<numComponents; j++) {
			gsl_matrix_set(subsetEigenVectorsMatrix, i, j, gsl_matrix_get(eigenVectors, i, j)); // TODO quicker copying
		}
	}

	gsl_matrix *pcaData = gsl_matrix_alloc(numComponents, numSpikes);
	gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1, subsetEigenVectorsMatrix, mdev, 0, pcaData);

	InputDataset *outData = new InputDataset(numComponents, numSpikes, 1);

#pragma omp parallel for
	for(int i=0; i < numComponents; i++) {
		for(int j=0; j < numSpikes; j++) {
			outData->data[j][i] = gsl_matrix_get(pcaData, i, j);
		}
	}

	gsl_matrix_free(pcaData);
	gsl_matrix_free(covarianceMatrix);
	gsl_matrix_free(subsetEigenVectorsMatrix);
	gsl_matrix_free(eigenVectors);
	gsl_vector_free(eigenValues);

	delete[] mean;
	return outData;
}

InputDataset* PcaReduction::reverse(const InputDataset *input) {
	return NULL;
}
