#include "waveletreductionks.h"

#include <assert.h>
#include <string.h>
#include <iostream>
#include <math.h>
#include <algorithm>
#include <gsl/gsl_cdf.h>
#include <list>
#include <map>

WaveletReductionKS::WaveletReductionKS()
{
	outDimensions = 10;
	waveletType = gsl_wavelet_haar;
	k = 2;
}

WaveletReductionKS::~WaveletReductionKS()
{

}

InputDataset* WaveletReductionKS::reduce(const InputDataset *input) {
	return reduce(input, false);
}

InputDataset* WaveletReductionKS::reduce(const InputDataset *input, bool manualIndexes)
{
	int inputSampleSize = input->data[0].size();
	n = 1<< (int)ceil(log2(inputSampleSize)); // pow2-padded

	InputDataset *outData = new InputDataset(outDimensions, input->nSamples, 1);
	outData->resizeVectorsToParameters();

	gsl_wavelet *wavelet = gsl_wavelet_alloc(waveletType, k);
	gsl_wavelet_workspace *work = gsl_wavelet_workspace_alloc(n);



	assert(wavelet != NULL);
	assert(work != NULL);

	double *data = new double[n];

	std::vector<std::vector<double> > transformedData(  input->nSamples, std::vector<double>( n, 0));
	std::vector<double> means(n);
	std::vector<double> variances(n);
	std::vector<double> distributionDifferences(n);

	//TODO could be parallelized with some effort
	for(unsigned int sampleIndex=0;sampleIndex<input->nSamples;sampleIndex++) {
		memset(data, 0, n*sizeof(double));
		std::copy(input->data[sampleIndex].begin(), input->data[sampleIndex].end(), data);

		gsl_wavelet_transform_forward(wavelet, data, 1, n, work);
		for(int i=0;i < n; i++) {
			transformedData[sampleIndex][i] = data[i];
			//std::cout << data[i] << " ";
		}
		//std::cout << std::endl;
	}

	// means
	//std::cout << "means: ";
	for(int dim=0; dim < n; dim++) {
		double mean = 0;
		for(unsigned int sample=0; sample < input->nSamples; sample++) {
			mean += transformedData[sample][dim];
		}
		mean /= input->nSamples;
		means[dim] = mean;
		//std::cout << mean << " ";
	}
	//std::cout << std::endl;

	// variances
	for(int dim=0; dim < n; dim++) {
		double variance = 0;
		for(unsigned int sample=0; sample < input->nSamples; sample++) {
			variance += pow(transformedData[sample][dim] - means[dim], 2);
		}
		variances[dim] = variance / input->nSamples;
	}

	std::vector<double> sortedSamples(input->nSamples);
	std::vector<double> probabilities(input->nSamples);

	// Empirical CDF and KS distance
	for(int dim=0; dim < n; dim++) {

		for(unsigned int sample=0; sample < input->nSamples; sample++) {
			sortedSamples[sample] = transformedData[sample][dim];
		}
		std::sort(sortedSamples.begin(), sortedSamples.end());


		for(unsigned int i=0; i< sortedSamples.size()-1;) {
			int forwardCounter = 0;
			while(sortedSamples[i] == sortedSamples[i + forwardCounter + 1] && (i + forwardCounter + 1) < sortedSamples.size()-1) {
				forwardCounter++;
			}
			int maxIndex = i + forwardCounter +1;
			for(int j=i;j < maxIndex; j++) {
				probabilities[j] = ((double)maxIndex) / sortedSamples.size();
			}
			probabilities[sortedSamples.size()-1] = 1;
			i = maxIndex;
		}

		double distributionDifference = 0;
		if(means[dim] == 0 || variances[dim] < 0.0010) {
			distributionDifferences[dim] = 0;
		} else {
			for(std::vector<double>::iterator it = sortedSamples.begin(); it != sortedSamples.end(); it++) {
				double value = *it;
				double z = (value-means[dim])/sqrt(variances[dim]);
				int probabilityIndex = std::lower_bound(sortedSamples.begin(), sortedSamples.end(), value) - sortedSamples.begin();
				double localDistributionDifference = fabs(probabilities[probabilityIndex] - gsl_cdf_ugaussian_P(z));
				if(localDistributionDifference > distributionDifference) {
					distributionDifference = localDistributionDifference;
				}
			}

			distributionDifferences[dim] = distributionDifference;
		}
	}

	if(!manualIndexes) {
		indexes = new int[n];
		double* devData = new double[n];

		std::copy(distributionDifferences.begin(), distributionDifferences.end(), devData);

		for(int i=0; i < n; i++) {
			indexes[i] = i;
		}

		//insertion sort
		for(int j = 1; j < n; j++){

			double key = devData[j];
			int keyIndex = indexes[j];
			int i = j - 1;

			while(i >= 0 && devData[i] > key){
				devData[i + 1] = devData[i];
				indexes[i + 1] = indexes[i];
				i--;
			}
			devData[i + 1] = key;
			indexes[i + 1] = keyIndex;
		}


	} else {
		std::cout << "Used manual indexes: ";
	}
	// Print the indexes
	std::cout << "KS wavelet\nindexes= {";
	for(int i=0; i < n-1; i++) {
		std::cout << indexes[i] << ", ";
	}
	std::cout << indexes[n-1] << "};" << std::endl;

	// Now sorted
	for(unsigned int dimIndex=0; dimIndex < outDimensions; dimIndex++) {
		for(unsigned int spike=0; spike < input->nSamples; spike++) {
			outData->data[spike][dimIndex] = transformedData[spike][indexes[n-1-dimIndex]];
		}
	}

	// outData->data[sampleIndex][i] = data[i];

	gsl_wavelet_free (wavelet);
	gsl_wavelet_workspace_free (work);
	delete[] data;

	outData->isReduced = true;
	outData->reduceClass = this;
	return outData;
}

InputDataset* WaveletReductionKS::reverse(const InputDataset *input) {
	int outN = ((WaveletReductionKS*) input->reduceClass)->n; // TODO maybe has to be set to 200 manually...

	InputDataset *outData = new InputDataset(outN, input->nSamples, input->nChannels);
	outData->resizeVectorsToParameters();

	gsl_wavelet *wavelet = gsl_wavelet_alloc(waveletType, k);
	gsl_wavelet_workspace *work = gsl_wavelet_workspace_alloc(n);

	assert(wavelet != NULL);
	assert(work != NULL);

	double *data = new double[outN];

	//TODO could be parallelized with some effort
	for(unsigned int sampleIndex=0;sampleIndex<input->nSamples;sampleIndex++) {
		memset(data, 0, n*sizeof(double));
		for(unsigned int i=0; i < outDimensions; i++) {
			data[indexes[n-1-i]] = input->data[sampleIndex][i];
		}

		gsl_wavelet_transform_inverse(wavelet, data, 1, n, work);
		for(int i=0;i < outN; i++) {
			outData->data[sampleIndex][i] = data[i];
		}
	}

	gsl_wavelet_free (wavelet);
	gsl_wavelet_workspace_free (work);
	delete[] data;
	outData->reduceClass = this;
	return outData;
}
