#include <boost/timer.hpp>
#include <iostream>
#include <float.h>
#include <assert.h>
#include <set>

#include "kmeans.h"

#ifdef PROFILEKM
#include "google/profiler.h"
#endif

using namespace tbb;

#define Uint32 unsigned int

KMeans::KMeans(std::vector<std::vector<double> > points, int numCentroids, int cutoff):
	centroids(NULL),
	rawCentroids(NULL)
{
	this->points = points;
	this->numCentroids = numCentroids;
	this->nDimensions = points[0].size();
	this->numPoints = points.size();
	this->cutoff = cutoff;

	convergenceStatus = 0;
	numIterationsRun = 0;

	clusterIds.resize(numPoints);

//	arrayPoints = new double[numPoints * nDimensions];
//	for(unsigned int point = 0; point < numPoints; ++point) {
//		for(unsigned int dimension = 0; dimension < nDimensions; ++dimension) {
//			arrayPoints[point * nDimensions + dimension] = this->points[point][dimension];
//		}
//	}

	initCentroids();	
}


KMeans::~KMeans() {
	if(rawCentroids)
		delete[] rawCentroids;

	if(centroids)
		delete[] centroids;

//	if(arrayPoints)
//		delete[] arrayPoints;
}




int KMeans::loopSerial() {
	int numUpdatedPoints = 0;

	float *tempNewClusters = new float[numCentroids * nDimensions];
	Uint32 *tempNewClusterSize =  new Uint32[numCentroids];
	memset(tempNewClusters, 0, sizeof(float) * numCentroids * nDimensions);
	memset(tempNewClusterSize, 0, sizeof(Uint32) * numCentroids);

	classifyPoints(0, numPoints, &numUpdatedPoints, tempNewClusters, tempNewClusterSize);


	for(int centroid = 0; centroid < numCentroids; ++centroid)
	  for(int dimension = 0; dimension < nDimensions; ++dimension)
		centroids[centroid].pos[dimension] = tempNewClusters[centroid*nDimensions + dimension] / tempNewClusterSize[centroid];

	convergenceStatus = ((float)numUpdatedPoints/(float)numPoints < CONVERGENCE_DELTA);

	delete[] tempNewClusters;
	delete[] tempNewClusterSize;

	return ++numIterationsRun;
}

int KMeans::loopParallel() {
	int numUpdatedPoints = 0;

	float *tempNewClusters = new float[numCentroids * nDimensions];
	Uint32 *tempNewClusterSize =  new Uint32[numCentroids];
	memset(tempNewClusters, 0, sizeof(float) * numCentroids * nDimensions);
	memset(tempNewClusterSize, 0, sizeof(Uint32) * numCentroids);

	KMeansTask &a = *new (task::allocate_root()) KMeansTask(this, 0, numPoints, &numUpdatedPoints, this->cutoff, tempNewClusters, tempNewClusterSize);
	KMeansTask::spawn_root_and_wait(a);

	convergenceStatus = ((float)numUpdatedPoints/(float)numPoints < CONVERGENCE_DELTA);

	for(int centroid = 0; centroid < numCentroids; ++centroid)
	  for(int dimension = 0; dimension < nDimensions; ++dimension)
		centroids[centroid].pos[dimension] = tempNewClusters[centroid*nDimensions + dimension] / tempNewClusterSize[centroid];

	delete[] tempNewClusters;
	delete[] tempNewClusterSize;

	return ++numIterationsRun;
}

void KMeans::OnLoop() {
	//static int frame = 0;
	static boost::timer time;

	if(convergenceStatus == 2)
	  return;

	if(!numIterationsRun)
		time.restart();

	if(convergenceStatus == 0) {
		numIterationsRun = loopParallel();
		//numIterationsRun = loopSerial();
	} else if (convergenceStatus == 1) {
		std::cout << "Converged in " << numIterationsRun << " iterations." << std::endl;
		std::cout << "Time taken: " << time.elapsed() << " secs.\n" << std::endl;
		convergenceStatus = 2;
	}
}

void KMeans::convertToVectorFormat() {
	lastComputedClusters.clear();
	lastComputedClusters.resize(numCentroids);
	for(unsigned int point = 0; point < numPoints; ++point) {
		lastComputedClusters[clusterIds[point]].push_back(point);
	}
}


void KMeans::classifyPoints(int start, int count, int *pointsUpdated, float *tempNewClusters, Uint32 *tempNewClusterSize ) {
	Uint32 point = start;
	int dimension, closestCentroid, centroid;
	float distance, tempDistance;

	//tempNewClusters = new float[numCentroids * numDimensions];
	//tempNewClusterSize = new Uint32[numCentroids];

	//memset(tempNewClusters, 0, sizeof(float) * numCentroids * numDimensions);
	//memset(tempNewClusterSize, 0, sizeof(Uint32) * numCentroids);

	// assign each point to its nearest centroid
	//for (point = start; point < (start+count); ++point) {
	for(std::vector<std::vector<double> >::const_iterator it = points.begin()+start; it != points.begin()+start+count; ++it, ++point) {
		distance = FLT_MAX;
		closestCentroid = -1;
		for (centroid = 0; centroid < numCentroids; ++centroid) {
			tempDistance = 0;
			for (dimension = 0; dimension < nDimensions; ++dimension)
				//TODO: abort if tempD > dist?
				tempDistance += ((*it)[dimension] - centroids[centroid].pos[dimension]) * ((*it)[dimension] - centroids[centroid].pos[dimension]);
				//tempDistance += (arrayPoints[point*nDimensions + dimension] - centroids[centroid].pos[dimension]) * (arrayPoints[point*nDimensions + dimension] - centroids[centroid].pos[dimension]);

			if (tempDistance < distance) {
				closestCentroid = centroid;
				distance = tempDistance;
			}
		}
		if(clusterIds[point] != closestCentroid) {
			clusterIds[point] = closestCentroid;
			++(*pointsUpdated);
		}
		for (dimension = 0; dimension < nDimensions; ++dimension) {
		  tempNewClusters[closestCentroid * nDimensions + dimension] += (*it)[dimension];
		  //tempNewClusters[closestCentroid * nDimensions + dimension] += arrayPoints[point*nDimensions + dimension];
		}
		++tempNewClusterSize[closestCentroid];

		assert(clusterIds[point] > -1);
	}

	// average each centroid
	//for(centroid = 0; centroid < numCentroids; ++centroid)
	//  for (dimension = 0; dimension < numDimensions; ++dimension)
	//    centroids[centroid].pos[dimension] = tempNewClusters[centroid*numDimensions + dimension] / tempNewClusterSize[centroid];

	//delete[] tempNewClusters;
	//delete[] tempNewClusterSize;

}

//FIXME: init them to the bounds of the dataset, not 640/480
void KMeans::initCentroids() {
	float *color;

	if(rawCentroids)
		delete[] rawCentroids;
	rawCentroids = new float[numCentroids * nDimensions];

	if(centroids)
		delete[] centroids;
	centroids = new Centroid[numCentroids];

	for(int centroid = 0; centroid < numCentroids; ++centroid)
		centroids[centroid].pos = &rawCentroids[centroid*nDimensions];

	std::set<int> takenPoints;
	for (int centroid = 0; centroid < numCentroids; ++centroid) {
		//uncomment the following to get a random init
		int index = rand() % numPoints; //TODO: terrible way of doing it..

		//don't init >1 cluster to the same point
		while (find(takenPoints.begin(), takenPoints.end(), index) != takenPoints.end())
		  index = rand() % numPoints;

		takenPoints.insert(index);

		for (int dimension = 0; dimension < nDimensions; ++dimension) {
			// the first inits a random place, the second on a random point.
			// the latter guarantees that every centroid will have at least one point assigned, GIVEN that each point is unique...
			//centroids[centroid].pos[dimension] = ((rand() % 1000) / 1000.0f) * (maxPoints[dimension] - minPoints[dimension]);
			centroids[centroid].pos[dimension] = points[index][dimension];
		}

		//std::cout << "Centroid #" << i << " at " << index << std::endl;
		//std::cout << "Pos: ";
		//for (int dimension = 0; dimension < nDimensions; ++dimension) {
			//std::cout << centroids[centroid]
		//}


		color = centroids[centroid].color;
		switch (centroid) {

			case 0:
				color[0] = 1.0f;
				color[1] = 1.0f;
				color[2] = 1.0f;
				break;
			case 1:
				color[0] = 1.0f;
				color[1] = 0.0f;
				color[2] = 1.0f;
				break;
			case 2:
				color[0] = 0.0f;
				color[1] = 1.0f;
				color[2] = 1.0f;
				break;
			case 3:
				color[0] = 0.0f;
				color[1] = 0.0f;
				color[2] = 1.0f;
				break;
			case 4:
				color[0] = 1.0f;
				color[1] = 0.0f;
				color[2] = 0.0f;
				break;
			case 5:
				color[0] = 0.0f;
				color[1] = 1.0f;
				color[2] = 0.0f;
				break;
			default:
				color[0] = (rand() % 100) / 100.0f;
				color[1] = (rand() % 100) / 100.0f;
				color[2] = (rand() % 100) / 100.0f;
		}
	}
}


KMeansTask::KMeansTask(KMeans *_km, int _start, int _numPoints, int *_numUpdatedPoints, int _serialThreshold, float *_tempNewClusters, Uint32 *_tempNewClusterSize) :
	serialThreshold(_serialThreshold),
	numPoints(_numPoints),
	numUpdatedPoints(_numUpdatedPoints),
	start(_start),
	km(_km),
	tempNewClusters(_tempNewClusters),
	tempNewClusterSize(_tempNewClusterSize)
{
}

task *KMeansTask::execute() {
#ifdef PROFILEKM
	ProfilerRegisterThread();
#endif
	if(numPoints < serialThreshold) {
		km->classifyPoints(start, numPoints, numUpdatedPoints, tempNewClusters, tempNewClusterSize);
		return NULL;
	}

	int splitNumPoints = numPoints / 2;

	int numUpdatedPointsTask1 = 0;
	int numUpdatedPointsTask2 = 0;

	int size = km->numCentroids * km->nDimensions;

	//TODO: is this kosher? i really don't like allocating memory like this, but i don't
	//know how many times this will be called because of the way it parallelizes
	float *tempNewClusters1 = new float[size];
	Uint32 *tempNewClusterSize1 =  new Uint32[km->numCentroids];
	memset(tempNewClusters1, 0, sizeof(float) * size);
	memset(tempNewClusterSize1, 0, sizeof(Uint32) * km->numCentroids);

	float *tempNewClusters2 = new float[size];
	Uint32 *tempNewClusterSize2 =  new Uint32[km->numCentroids];
	memset(tempNewClusters2, 0, sizeof(float) * size);
	memset(tempNewClusterSize2, 0, sizeof(Uint32) * km->numCentroids);

	KMeansTask &subTask1 = *new (allocate_child()) KMeansTask(this->km, start, splitNumPoints, &numUpdatedPointsTask1, serialThreshold, tempNewClusters1, tempNewClusterSize1);

	int offset = splitNumPoints;
	if(numPoints % 2)
	  offset += 1;

	KMeansTask &subTask2 = *new (allocate_child()) KMeansTask(this->km, start+splitNumPoints, offset, &numUpdatedPointsTask2, serialThreshold, tempNewClusters2, tempNewClusterSize2);

	set_ref_count(3);
	spawn(subTask2);

	spawn_and_wait_for_all(subTask1);

	*numUpdatedPoints = numUpdatedPointsTask1 + numUpdatedPointsTask2;

	for(int centroid = 0; centroid < km->numCentroids; ++centroid) {
	  for(int dimension = 0; dimension < km->nDimensions; ++dimension) {
		int index = centroid * km->nDimensions + dimension;
		tempNewClusters[index] = tempNewClusters1[index] + tempNewClusters2[index];
	  }
	  tempNewClusterSize[centroid] = tempNewClusterSize1[centroid] + tempNewClusterSize2[centroid];
	}

	delete[] tempNewClusters1;
	delete[] tempNewClusters2;
	delete[] tempNewClusterSize1;
	delete[] tempNewClusterSize2;

	return NULL;
}
