#include <stdio.h>
#include <vector>
#include <stack>
#include <math.h>
#include <set>
#include <ctime>

#include "spc.h"
#include "clustercollection.h"

#include "gmst.hpp"
#include "sfcnn_knng.hpp"
#include "dpoint.hpp"




#ifdef PROFILESPC
#include "google/profiler.h"
#endif

#include <tbb/tick_count.h>

#define NDEBUG
#ifndef NDEBUG
#define D(a) std::cout << (a) << std::endl;
#else
#define D(a) ;
#endif

#ifndef MST_DIMENSION
#define MST_DIMENSION 16
#endif

tbb::mutex printMutex;

typedef struct beginend {
   const int a, b;
   int begin() const { return a; }
   int end() const { return b; }
} beginend;



#define SERIAL_FOR_START(start, end )  beginend r = {start, end};
#define SERIAL_FOR_END()




Spc::Spc(std::vector<std::vector<double> > points) {
	unsigned int nNearestNeighbors = 11;

	if(points.size() <= nNearestNeighbors) {
		nNearestNeighbors = points.size() - 1; // can max have nPoints - itself num of neighbors
		std::cout << "Reducing number of neighbors to " << nNearestNeighbors << " because the dataset is too small." << std::endl;
	}

	thetaForGij = (nSpinStates*theta-1)/(nSpinStates-1);
	this->points = points;
	nPoints = (unsigned int) points.size();
	nDimensions = (unsigned int) points[0].size();

	clusterIds.resize(nPoints);

	susceptibilityPlot = new XyPlot();
	correlationClusterSizesPlot = new XyCollectionPlot();
	shouldCancelSusceptibilityPlot = false;

	randomGen = new RandomGenerator(0, nSpinStates-1);

	buildKdtree();
	D("built tree");
	determineClosestNeighborsAndInteractionStrengths(nNearestNeighbors);
	D("found neighbors");

	initialGuessTemperature = round(1000.0 * exp(-0.5) / (4*log(1 + sqrt(nSpinStates)))) / 1000.0;
	std::cout << "Guess temp: " << initialGuessTemperature << std::endl;
}

Spc::~Spc() {
	delete[] kdPoints;
	delete kdTree;
	delete randomGen;
}

void Spc::printPoints() {
	for (int i = 0; i < nPoints; i++) {
		std::cout << "points[" << i << "]: ";
		for (int j = 0; j < nDimensions; j++) {
			std::cout << points[i][j] << " ";
		}
		std::cout << std::endl;
	}
}

void Spc::buildKdtree() {
	kdPoints = annAllocPts(nPoints, nDimensions); // data points
	for (int i = 0; i < nPoints; i++) {
		for (int j = 0; j < nDimensions; j++) {
			kdPoints[i][j] = points[i][j];
		}
	}

	kdTree = new ANNkd_tree(kdPoints, nPoints, nDimensions);
}

void Spc::determineClosestNeighborsAndInteractionStrengths(unsigned int nNearestNeighbors) {
	searchKdTreeAndCopyNeighborData(nNearestNeighbors);
	removeImmutualNeighborsAndBackNeighbors();
	addMstToNeighborGraph(); // comment to remove mst
	checkConnectedNeighborGraph();

	// Determine average neighbor-distance (a)
	unsigned int pairs = 0;
	double avgDistance = 0.0;
	for (int i = 0; i < nPoints; i++) {
		pairs += (unsigned int) (((neighborDistances[i].size())));
		for (unsigned int j = 0; j < neighborDistances[i].size(); j++) {
			avgDistance += neighborDistances[i][j];
		}
	}

	avgDistance /= pairs;
	calculateInteractionStrengths(pairs, avgDistance);

}

void Spc::searchKdTreeAndCopyNeighborData(unsigned int nNearestNeighbors) {
	tbb::tick_count start = tbb::tick_count::now();

	unsigned int K1 = nNearestNeighbors + 1; // search for one more, and then remove self (in the rare case of overlapping points)

	neighborIndexes.resize(nPoints);
	neighborDistances.resize(nPoints);
	interactionStrengths.resize(nPoints);


	SERIAL_FOR_START(0, nPoints); // this is to make it easier to parallelize when thread safe implementation is used.
	ANNidx *kdIndexes = new ANNidx[K1]; // WORK
	ANNdist *kdSquareDistances = new ANNdist[K1]; // WORK
	for (int i = r.begin(); i != r.end(); i++) {
		ANNpoint queryPoint = kdPoints[i];
		kdTree->annkSearch(queryPoint, K1, kdIndexes, kdSquareDistances, annErrorToleranceFraction);
		neighborIndexes[i].resize(nNearestNeighbors);
		neighborDistances[i].resize(nNearestNeighbors);
		int skip = 0;
		for (unsigned int j = 0; j < nNearestNeighbors; j++) {
			if (kdIndexes[j] == i) {
				skip = 1; // Used not to copy self reference
			}
			neighborIndexes[i][j] = kdIndexes[j + skip];
			neighborDistances[i][j] = sqrt(kdSquareDistances[j + skip]); // ||x-y||^2 == sum((xi-yi)^2), if we skip this we dont have to square at the interaction str
		}
	}
	delete[] kdIndexes;
	delete[] kdSquareDistances;
	SERIAL_FOR_END();

	tbb::tick_count stop = tbb::tick_count::now();
	std::cout << "time for Neighbor search: " << (stop - start).seconds() << std::endl;

}

void Spc::removeImmutualNeighborsAndBackNeighbors() {
	tbb::tick_count start = tbb::tick_count::now();
	for (int i = 0; i < nPoints; i++) {
		for (unsigned int j = 0; j < neighborIndexes[i].size(); j++) {
			int otherIndex = neighborIndexes[i][j];
			if (otherIndex < i) {
				// Remove back-neighbors
				neighborIndexes[i].erase(neighborIndexes[i].begin() + j);
				neighborDistances[i].erase(neighborDistances[i].begin() + j);
				j--;
			} else {
				// Remove immutual neighbors
				bool mutual = false;
				for (unsigned int k = 0; k < neighborIndexes[otherIndex].size(); k++) {
					if (neighborIndexes[otherIndex][k] == i) {
						mutual = true;
						break;
					}
				}

				if (!mutual) {
					neighborIndexes[i].erase(neighborIndexes[i].begin() + j);
					neighborDistances[i].erase(neighborDistances[i].begin() + j);
					j--;
				}
			}
		}
	}
	tbb::tick_count stop = tbb::tick_count::now();
	std::cout << " time for removeImmutualAndBackNeighbors: " << (stop - start).seconds() << std::endl;
}

void Spc::checkConnectedNeighborGraph() {

	ClusterCollection clusters(nPoints);
	for (int i = 0; i < nPoints; i++) {
		for (unsigned int j = 0; j < neighborIndexes[i].size(); j++) {
			int neighborIndex = neighborIndexes[i][j];

			if (!clusters.isAssignedToCluster(i) && !clusters.isAssignedToCluster(neighborIndex)) {
				int clusterIndex = clusters.startCluster();
				clusters.assign(clusterIndex, i);
				clusters.assign(clusterIndex, neighborIndex);
			} else if (clusters.isAssignedToCluster(i) && !clusters.isAssignedToCluster(neighborIndex)) {
				int clusterIndex = clusters.getContainingCluster(i);
				clusters.assign(clusterIndex, neighborIndex);
			} else if (!clusters.isAssignedToCluster(i) && clusters.isAssignedToCluster(neighborIndex)) {
				int clusterIndex = clusters.getContainingCluster(neighborIndex);
				clusters.assign(clusterIndex, i);
			} else if (clusters.getContainingCluster(i) != clusters.getContainingCluster(neighborIndex)) {
				clusters.markMergeClustersForPointIndexes(i, neighborIndex);
			}
		}
	}
	clusters.doMerge();
	clusters.removeEmptyClusters();

	std::cout << "Enforced connectedness. All points are connected in " << clusters.clusters.size() << " subgraphs:" << std::endl;

//	std::vector<std::vector<int> > cClusters = clusters.clusters;
//	std::cerr << "Printing " << cClusters.size() << " clusters.." << std::endl;
//	for(unsigned int cluster = 0; cluster < cClusters.size(); ++cluster) {
//	  std::cerr << "Cluster #" << cluster << " (" << cClusters[cluster].size() << "): ";
//	  std::sort(cClusters[cluster].begin(), cClusters[cluster].end());
//	  for(unsigned int point = 0; point < cClusters[cluster].size(); ++point) {
//		std::cerr << cClusters[cluster][point] << " ";
//	  }
//	  std::cerr << std::endl;
//	}
//	clusters.printClusters();

}

unsigned long long int doubleHash(double d, unsigned int leftRotate) {
	unsigned long long int hash = *(unsigned long long int*)(&d);
	return (hash << leftRotate) | (hash >> (64-leftRotate));
}

void Spc::addMstToNeighborGraph() {

	std::cout << "MST" << std::endl;
	if(nDimensions != MST_DIMENSION) {
		std::cout << "MST not added, as it is hard-coded for " << MST_DIMENSION << " dimensions" << std::endl;
		return;
	}

	tbb::tick_count startMst = tbb::tick_count::now();

	typedef reviver::dpoint<double,MST_DIMENSION> StannPoint;
	std::vector<StannPoint> stannPoints;
	stannPoints.resize(nPoints);
	std::set<unsigned long long int> uniqueXorSet; // Because STANN requires unique datapoints

	std::pair<std::set<unsigned long long int>::iterator, bool> ret;

	std::cout << "Copying into STANN datatype" << std::endl;
	for(int i=0;i<nPoints;i++) {
		unsigned long long int xorValue = 0;
		for(int currentDimension=0;currentDimension<nDimensions;currentDimension++) {
			xorValue ^= doubleHash(points[i][currentDimension], currentDimension);
		}
		ret = uniqueXorSet.insert(xorValue);

		while(ret.second == false) {
			unsigned int changeDim = nDimensions * randomGen->randFrac();
			std::cout << "Point " << i << " moved in dimension " << changeDim << std::endl;
			double currentValue = points[i][changeDim];
			xorValue ^= doubleHash(currentValue, changeDim); // revert old contribution
			currentValue += currentValue / 1e6 + 1e-10;
			xorValue ^= doubleHash(currentValue, changeDim); // add new contribution
			points[i][changeDim] = currentValue;
			ret = uniqueXorSet.insert(xorValue);
		}

		for(int j=0;j<nDimensions;j++) {
			stannPoints[i][j] = points[i][j];
		}

	}

	typedef std::vector<StannPoint>::size_type stype;

	std::vector< std::pair<stype,stype> > outputmst;

	std::cout << "Starting MST" << std::endl;
	gmst(stannPoints, outputmst);
	std::cout << "MST complete - adding edges" << std::endl;

	for(unsigned int i=0;i<outputmst.size();i++) {
		int from = outputmst[i].first;
		int to =  outputmst[i].second;
		if(from < to) {
			connectPoints(from, to);
		} else {
			connectPoints(to, from);
		}
	}

	tbb::tick_count endMst = tbb::tick_count::now();
	std::cout << "time for MST: " << (endMst-startMst).seconds() << std::endl;
}

void Spc::connectPoints(int fromIndex, int toIndex) {
	if (std::find(neighborIndexes[fromIndex].begin(), neighborIndexes[fromIndex].end(), toIndex) == neighborIndexes[fromIndex].end()) {
		neighborIndexes[fromIndex].push_back(toIndex);
		neighborDistances[fromIndex].push_back(distance(fromIndex, toIndex));
		//std::cerr << "Connected " << fromIndex << ", " << toIndex << std::endl;
	} else {
		//std::cerr << "Did not connect " << fromIndex << ", " << toIndex << " because they were already in the same cluster" << std::endl;
	}
}

double Spc::distance(int fromIndex, int toIndex) {
	double distance = 0;
	for(unsigned int i=0;i<points[fromIndex].size();i++) {
		distance += pow(points[fromIndex][i] - points[toIndex][i],2.0);
	}
	return sqrt(distance);
}

void Spc::calculateInteractionStrengths(unsigned int pairs, double avgDistance) {
	tbb::tick_count start = tbb::tick_count::now();
	double kHat = (2.0 * pairs) / nPoints; // Blatt section 4.1.3

	for (int i = 0; i < nPoints; i++) {
		interactionStrengths[i].resize(neighborIndexes[i].size());
		for (unsigned int j = 0; j < neighborDistances[i].size(); j++) {
			double Jij = (1.0 / kHat) * exp(-(neighborDistances[i][j] * neighborDistances[i][j]) / (2.0 * avgDistance * avgDistance)); // Blatt eq 4.1
			interactionStrengths[i][j] = Jij;
		}
	}
	tbb::tick_count stop = tbb::tick_count::now();
	std::cout << "time for calculateInteractionStrengths: " << (stop - start).seconds() << std::endl;
}

XyPlot* Spc::plotSusceptibility(double minTemperature, double maxTemperature, double step) {
#ifdef PROFILESPC
	ProfilerStart("profile/plotSusceptibility");
#endif
	tbb::tick_count start = tbb::tick_count::now();
	D("Susceptibility");
	int numResults = (int) ((maxTemperature - minTemperature) / step);

	int *iterationsPerformed = new int(0);

	//tbb::task_scheduler_init init(1);

	tbb::parallel_for(tbb::blocked_range<size_t>(0,numResults), PlotSusceptibility(this, iterationsPerformed, minTemperature, step, numResults));

	delete iterationsPerformed;

	shouldCancelSusceptibilityPlot = false;
#ifdef PROFILESPC
	ProfilerStop();
#endif
	tbb::tick_count stop = tbb::tick_count::now();
	std::cout << "time for plotSusceptibility: " << (stop - start).seconds() << std::endl;
	return susceptibilityPlot;
}

double Spc::calculateSusceptibility(const double temperature, bool determineSpinSpinCorrelation, RandomGenerator *innerRandom, bool printClusteringResults, int *spins) {
	if(innerRandom == NULL)
		innerRandom = randomGen;

	std::vector<std::vector<double> > probabilityValues = calculateProbabilityValues(temperature);

	bool spinsWereSpecifiedExternally = true;

	if(spins == NULL) { // Local spins if not reusing from outer scope
		spins = new int[nPoints];
		spinsWereSpecifiedExternally = false;
		// Assign random initial spin
		for (int i = 0; i < nPoints; i++) {
			spins[i] = innerRandom->randInRange();
		}
	}

	double magnetizationSum = 0.0;
	double magnetizationSquareSum = 0.0;

	ClusterCollection clusters = ClusterCollection(nPoints);
	int spinCount[nSpinStates];

	// Setup correlation sum
	std::vector<std::vector<double> > correlationSum(nPoints); // can be int until averaging
	for (unsigned int i = 0; i < correlationSum.size(); i++) {
		correlationSum[i].resize(neighborIndexes[i].size(), 0.0);
	}

	for (int iteration = 0; iteration < nMonteCarloIterations + nWarmupIterations; iteration++) {
		// Empty clusters
		clusters.reset();

		for (int i = 0; i < nPoints; i++) {
			for (unsigned int j = 0; j < neighborIndexes[i].size(); j++) {
				int neighborIndex = neighborIndexes[i][j];

				if (spins[neighborIndex] == spins[i] && probabilityValues[i][j] > innerRandom->randFrac()) {
					if (!clusters.isAssignedToCluster(i) && !clusters.isAssignedToCluster(neighborIndex)) {
						int clusterIndex = clusters.startCluster();
						clusters.assign(clusterIndex, i);
						clusters.assign(clusterIndex, neighborIndex);
					} else if (clusters.isAssignedToCluster(i) && !clusters.isAssignedToCluster(neighborIndex)) {
						int clusterIndex = clusters.getContainingCluster(i);
						clusters.assign(clusterIndex, neighborIndex);
					} else if (!clusters.isAssignedToCluster(i) && clusters.isAssignedToCluster(neighborIndex)) {
						int clusterIndex = clusters.getContainingCluster(neighborIndex);
						clusters.assign(clusterIndex, i);
					} else if (clusters.getContainingCluster(i) != clusters.getContainingCluster(neighborIndex)) {
						clusters.markMergeClustersForPointIndexes(i, neighborIndex);
					}
				}
			}
		}

		clusters.doMerge();

		// Change spins
		for (int i = 0; i < nPoints; i++) {
			spins[i] = innerRandom->randInRange();
		}
		for (unsigned int i = 0; i < clusters.clusters.size(); i++) {
			std::vector<int> *currentCluster = &clusters.clusters[i];
			int newSpin = innerRandom->randInRange();
			for (std::vector<int>::const_iterator it = currentCluster->begin(); it != currentCluster->end(); ++it) {
				spins[*it] = newSpin;
			}
		}

		if (iteration >= nWarmupIterations) {

			// Spin-spin correlation
			if (determineSpinSpinCorrelation) {
				for (int i = 0; i < nPoints; i++) {
					for (unsigned int j = 0; j < correlationSum[i].size(); j++) {
						if (spins[i] == spins[neighborIndexes[i][j]]) {
							correlationSum[i][j] += 1.0;
						}
					}
				}
			}

			// Count number of spin members
			memset(spinCount, 0, nSpinStates * sizeof(int));
			for (int i = 0; i < nPoints; i++) {
				spinCount[spins[i]]++;
			}

			// Calculate magnetic properties
			int Nmax = 1; // Number of points in "biggest" spinState
			for (int i = 0; i < nSpinStates; i++) {
				if (spinCount[i] > Nmax) {
					Nmax = spinCount[i];
				}
			}
			//double currentEnergy = calculateEnergy(spins);
			double currentMagnetization = ((double) nSpinStates * Nmax - nPoints) / ((nSpinStates - 1.0) * nPoints); // Blatt eq 2.4
			magnetizationSum += currentMagnetization;
			magnetizationSquareSum += currentMagnetization * currentMagnetization;
			//std::cout << "Energy in iteration " << iteration << ": " << currentEnergy << std::endl;
		}
	}
	if(!spinsWereSpecifiedExternally) {
		delete[] spins;
	}
	double averageMagnetization = magnetizationSum / nMonteCarloIterations;
	double averageSquareMagnetization = magnetizationSquareSum / nMonteCarloIterations;

	double susceptibility = (double) nPoints / temperature * (averageSquareMagnetization - averageMagnetization	* averageMagnetization); // Blatt eq 2.6

	if (determineSpinSpinCorrelation) {
		for (int i = 0; i < nPoints; i++) {
			for (unsigned int j = 0; j < correlationSum[i].size(); j++) {
				correlationSum[i][j] /= nMonteCarloIterations;
			}
		}
		assert(correlationSum.size() == neighborIndexes.size());
		// Perform clustering based on spin-spin correlation
		ClusterCollection correlationClusters = ClusterCollection(nPoints);		
		for (int i = 0; i < nPoints; i++) {
			for (unsigned int j = 0; j < neighborIndexes[i].size(); j++) {
				int neighborIndex = neighborIndexes[i][j];

				if (correlationSum[i][j] > thetaForGij) {
					if (!correlationClusters.isAssignedToCluster(i) && !correlationClusters.isAssignedToCluster(neighborIndex)) {
						int clusterIndex = correlationClusters.startCluster();
						correlationClusters.assign(clusterIndex, i);
						correlationClusters.assign(clusterIndex, neighborIndex);
					} else if (correlationClusters.isAssignedToCluster(i) && !correlationClusters.isAssignedToCluster(neighborIndex)) {
						int clusterIndex = correlationClusters.getContainingCluster(i);
						correlationClusters.assign(clusterIndex, neighborIndex);
					} else if (!correlationClusters.isAssignedToCluster(i) && correlationClusters.isAssignedToCluster(neighborIndex)) {
						int clusterIndex = correlationClusters.getContainingCluster(neighborIndex);
						correlationClusters.assign(clusterIndex, i);
					} else if (correlationClusters.getContainingCluster(i) != correlationClusters.getContainingCluster(neighborIndex)) {
						correlationClusters.markMergeClustersForPointIndexes(i, neighborIndex);
					}
				}
			}
		}
		correlationClusters.doMerge();
		correlationClusters.removeEmptyClusters();

		std::vector<std::vector<int> > *clusters = &correlationClusters.clusters;


		clusterIds.resize(nPoints, -1);
		for (unsigned int cluster = 0; cluster < lastComputedClusters.size(); ++cluster) {
			for(unsigned int point = 0; point < lastComputedClusters[cluster].size(); ++point) {
				clusterIds[ lastComputedClusters[cluster][point] ] = cluster;
			}
		}

		if(printClusteringResults) {
		  std::cout << "Printing " << clusters->size() << " clusters.." << std::endl;
		  for(unsigned int cluster = 0; cluster < clusters->size(); ++cluster) {
			std::cout << "Cluster #" << cluster << " (" << (*clusters)[cluster].size() << "): ";
			std::sort((*clusters)[cluster].begin(), (*clusters)[cluster].end());
			for(unsigned int point = 0; point < (*clusters)[cluster].size(); ++point) {
			  std::cout << (*clusters)[cluster][point] << " ";
			}
			std::cout << std::endl;
		  }
		}

		std::vector<int> currentSizes;

		for(unsigned int i=0;i<clusters->size();i++) {
			currentSizes.push_back((*clusters)[i].size());
		}
		std::sort(currentSizes.begin(), currentSizes.end(), std::greater<int>());
		correlationClusterSizesPlot->insert(temperature, currentSizes);

		lastComputedClusters = *clusters;
	}
	if(temperature == 0.0) {
		return 0.0;
	}
	return susceptibility;
}

std::vector<std::vector<double> > Spc::calculateProbabilityValues(const double & temperature) {
	// Determine probabilities for temperature
	std::vector<std::vector<double> > probabilityValues = std::vector<std::vector<double> >(nPoints);
	for (int i = 0; i < nPoints; i++) {
		probabilityValues[i].resize(interactionStrengths[i].size());
		for (unsigned int j = 0; j < interactionStrengths[i].size(); j++) {
			if(temperature == 0)
				probabilityValues[i][j] = 1.0;
			else
				probabilityValues[i][j] = 1.0 - exp(-interactionStrengths[i][j] / temperature); // Blatt eq 3.2
		}
	}

	return probabilityValues;
}

double Spc::calculateEnergy(int *spins) {
	double energy = 0;
	for (int currentIndex = 0; currentIndex < nPoints; currentIndex++) {
		for (unsigned int j = 0; j < neighborIndexes[currentIndex].size(); j++) {
			int neighborIndex = neighborIndexes[currentIndex][j];

			if (spins[currentIndex] == spins[neighborIndex]) {
				energy -= interactionStrengths[currentIndex][j];
			}
		}
	}
	return energy;
}

void Spc::cancelSusceptibilityPlot() {
	shouldCancelSusceptibilityPlot = true;
}

PlotSusceptibility::PlotSusceptibility(Spc *spc, int *iterationsPerformed, double minTemperature, double step, int numResults) {
	this->spc = spc;
	this->iterationsPerformed = iterationsPerformed;
	this->minTemperature = minTemperature;
	this->step = step;
	this->numResults = numResults;
}

void PlotSusceptibility::operator()(tbb::blocked_range<size_t> &r) const {
	{
		printMutex.lock();
		std::cout << "SPC task range: [" << r.begin() << "," << r.end()-1 << "] (" << (r.end() - r.begin()) << ")" << std::endl;
		printMutex.unlock();
	}
#ifdef PROFILESPC
	ProfilerRegisterThread();
#endif
	RandomGenerator *innerRandom = new RandomGenerator(spc->randomGen);
	int *spins = new int[spc->nPoints];
	// Assign random initial spin
	for (int i = 0; i < spc->nPoints; i++) {
		spins[i] = innerRandom->randInRange();
	}

	for (unsigned int i = r.begin(); i != r.end() && !spc->shouldCancelSusceptibilityPlot; i++) {
		double temperature = minTemperature + step * i;
		double susceptibility = spc->calculateSusceptibility(temperature, true, innerRandom, false, spins);
		spc->susceptibilityPlot->insert(temperature, susceptibility);
		spc->finishness = (double)((*iterationsPerformed)++)/numResults;
	}
	delete[] spins;
	delete innerRandom;
}
